@@ -104,6 +104,10 @@ def agnostic_abs(x):
     xp = GPU_CPU.get_numpy(x)
     return xp.abs(x)
 
+def agnostic_angle(x):
+    xp = GPU_CPU.get_numpy(x)
+    return xp.angle(x)
+
 def agnostic_argmax(x, axis=None):
     xp = GPU_CPU.get_numpy(x)
     return xp.argmax(x, axis=axis)
@@ -114,7 +118,17 @@ def agnostic_conj(x):
 
 def agnostic_diff(x, n=1, axis=-1):
     xp = GPU_CPU.get_numpy(x)
-    return xp.diff(x,n=n,axis=axis)
+    return xp.diff(x,n=n,axis=axis, append=0)
+
+
+def agnostic_unwrap(x, axis=-1):
+    xp = GPU_CPU.get_numpy(x)
+    return xp.unwrap(x,axis=axis)
+
+def agnostic_roll(x, shift, axis=None):
+    xp = GPU_CPU.get_numpy(x)
+    return xp.roll(x,shift,axis=axis)
+
 
 
 def agnostic_fft(x,axis=-1):
@@ -146,6 +160,17 @@ def agnostic_zeros(x,size_tuple,dtype=numpy.double):
     xp = GPU_CPU.get_numpy(x)
     return xp.zeros(size_tuple,dtype=dtype)
 
+def agnostic_rescale(x):
+    xp = GPU_CPU.get_numpy(x)
+    return (x-xp.amin(x))/(xp.amax(x)-xp.amin(x))
+
+def soft_threshold(x,threshold,direction='greater'):
+    if direction=='>':
+        x[x>threshold]=threshold
+    else:
+        x[x<threshold]=threshold
+    return x
+
     
 '''
   Extract the analytic signal for a give frequency within a
@@ -172,7 +197,7 @@ def agnostic_analytic_signal(x,fc,fs,window_length=64,axis=1):
     est = complex_estimator(xp,nx,fc,fs,window_length)
     est = agnostic_fft(est)
     if axis==1:
-        return xp.multiply(x,xp.expand_dims(est, axis=1))
+        return xp.multiply(x,xp.expand_dims(est, axis=0))
     return xp.multiply(x,est)
     
     
