@@ -30,10 +30,14 @@ DEFAULT = "default"
 DESC = "description"
 universal_arglist = {
     "args" :          {DEFAULT : None, DESC : "Arguments a passed through to the underlying math library"},
+    "coeffs" :        {DEFAULT : None, DESC : "Filter coefficients"},
     "xaxis" :         {DEFAULT : None, DESC : "A numpy vector of distances along the fibre"},
     "taxis" :         {DEFAULT : None, DESC : "A numpy vector of times associated with columns of data"},
     "directory_in" :  {DEFAULT : NONE, DESC : "The subdirectory from which to read data"},
     "directory_out" : {DEFAULT : NONE, DESC : "The subdirectory where results will be written"},
+    "direction"     : {DEFAULT : ">", DESC : "The direction for applying the threshold, > or <"},
+    "double_ended"  : {DEFAULT : -1,   DESC : "For handling double-ended fibre [-1=single-ended, 0=start-of-fibre half, 1=end-of-fibre half"},
+    "edge_order"    : {DEFAULT :  1,   DESC : "Gradient is calculatd using N-th order accurate differences at the boundaries"},
     "command_list" :  {                DESC : "A list of sub-commands collected in a single macro"},
     "train" :         {                DESC : "A dictionary containing training parameters for keras, e.g. { 'epochs' : 150, 'batch_size' : 10 }"},
     "func" :          {                DESC : "Either rms_from_fft or te_from_fft"}, 
@@ -62,6 +66,7 @@ universal_arglist = {
     "c" :             {DEFAULT : 0.0,     DESC : "The intercept of a linear transform (y = m*x + c)"},
     "data_style" :    {DEFAULT : NONE,  DESC : "A string identifier for the data inside the WITSML file"},
     "method" :        {DEFAULT : "lin_fit", DESC : "The method for curve fitting, see scipy.optimize.curve_fit"},
+    "mode" :          {DEFAULT : "constant", DESC : "Filter edge handling, see the scipy.ndimage documentation."},
     "moment" :        {DEFAULT : 1,     DESC : "The order of central moment, see scipy.stats.moment"},
     "max_velocity":   {DEFAULT : 1600,  DESC : "The maximum phase velocity"},
     "min_velocity":   {DEFAULT : 1400,  DESC : "The minimum phase velocity"},
@@ -242,6 +247,84 @@ class AbsCommand(BasicCommand):
     def execute(self):
         self._result = extra_numpy.agnostic_abs(self._previous.result())
 
+'''
+ AngleCommand : wrappers the numpy.angle() function
+'''
+class AngleCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Take the angle of the complex input"
+        return docs
+
+    def isGPU(self):
+        return True
+
+    def execute(self):
+        self._result = extra_numpy.agnostic_angle(self._previous.result())
+'''
+ RescaleCommand : rescales the data from 0 to 1
+'''
+class RescaleCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Rescale the data from 0 to 1"
+        return docs
+
+    def isGPU(self):
+        return True
+
+    def execute(self):
+        self._result = extra_numpy.agnostic_rescale(self._previous.result())
+'''
+ SoftThresholdCommand : applies a threshold greater or less
+'''
+class SoftThresholdCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._direction = jsonArgs.get('direction','>')
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Applies a soft threshold."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['direction'] }
+        return docs
+
+    def isGPU(self):
+        return True
+
+    def execute(self):
+        self._result = extra_numpy.soft_threshold(self._previous.result(),direction=self._direction)
+
+
+'''
+ UnwrapCommand : unwraps using the numpy.unwrap() function
+'''
+class UnwrapCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._axis = jsonArgs.get('axis',-1)
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Unwrap angles in the selected axis direction using numpy.unwrap"
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis'] }
+        docs['args']['axis']['default']=-1
+        return docs
+
+    def isGPU(self):
+        return True
+
+    def execute(self):
+        self._result = extra_numpy.agnostic_unwrap(self._previous.result(), axis=self._axis)
+
+
+
 '''
  ConjCommand : wrappers the numpy.conj() function
 '''
@@ -311,6 +394,7 @@ class KMeansCommand(BasicCommand):
     def __init__(self,command, jsonArgs):
         super().__init__(command, jsonArgs)
         self._args = jsonArgs
+        self._prevstack = jsonArgs.get('commands',[None])
 
     def docs(self):
         docs={}
@@ -320,7 +404,23 @@ class KMeansCommand(BasicCommand):
 
     def execute(self):
         n_clusters = self._args.get('n_clusters',10)
-        self._result = extra_numpy.kmeans_clustering(self._previous.result(),n_clusters=n_clusters)
+        if self._prevstack[0] is not None:
+            shape_out = self._previous.result().shape
+            for a in range(len(self._prevstack)):
+                test_shape = self._prevstack[a].result().shape
+                if (test_shape[0]*test_shape[1])<(shape_out[0]*shape_out[1]):
+                    shape_out=test_shape
+            xrows = shape_out[0]*shape_out[1]
+            xcols = len(self._prevstack)+1
+            data_big = numpy.zeros((xrows,xcols))
+            for a in range(xcols-1):
+                data_big[:,a]=self._prevstack[a].result().flatten()[:xrows]
+            data_big[:,-1]=self._previous.result().flatten()[:xrows]
+            result = extra_numpy.kmeans_clustering(data_big,n_clusters=n_clusters)
+            self._result = numpy.reshape(result,shape_out)
+        else:
+            # Single input...
+            self._result = extra_numpy.kmeans_clustering(self._previous.result(),n_clusters=n_clusters)
 
 
 
@@ -521,6 +621,7 @@ class HashCommand(BasicCommand):
 class ButterCommand(BasicCommand):
     def __init__(self,command, jsonArgs):
         super().__init__(command, jsonArgs)
+        self._args = jsonArgs
         self._order = jsonArgs.get('order',5)
         self._btype = jsonArgs.get('type','lowpass')
         self._ptype = jsonArgs.get('padtype','even')
@@ -574,6 +675,7 @@ class GaussianCommand(BasicCommand):
         self._tdir = jsonArgs.get('tsigma',1.0)
         self._xorder = jsonArgs.get('xorder',5)
         self._torder = jsonArgs.get('torder',5)
+        self._mode = jsonArgs.get('mode', 'constant')
 
     def docs(self):
         docs={}
@@ -582,7 +684,65 @@ class GaussianCommand(BasicCommand):
         return docs
 
     def execute(self):
-        self._result = ndimage.gaussian_filter(self._previous.result(), [self._xdir, self._tdir], order=[self._xorder, self._torder],mode='constant')
+        self._result = ndimage.gaussian_filter(self._previous.result(), [self._xdir, self._tdir], order=[self._xorder, self._torder],mode=elf._mode)
+
+'''
+ SobelCommand : Applies a Sobel edge detection filter using signal.ndarray.sobel()
+'''
+class SobelCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._axis = jsonArgs.get('axis',-1)
+        self._mode = jsonArgs.get('mode','constant')
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Applies a Sobel edge detection filter using signal.ndarray.sobel()."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','mode'] }
+        return docs
+
+    def execute(self):
+        self._result = ndimage.sobel(self._previous.result(),axis=self._axis, mode=self._mode)
+
+'''
+ ConvolveCommand : Applies the supplied convolution filter
+'''
+class ConvolveCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._coeffs = jsonArgs.get('coeffs',[[1,0,-1],[2,0,-2],[1,0,-1]])
+        self._mode = jsonArgs.get('mode','constant')
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Convolves the supplied filter with the data."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['coeffs','mode'] }
+        docs['args']['coeffs']['default'] = [[1,0,-1],[2,0,-2],[1,0,-1]] 
+        return docs
+
+    def execute(self):
+        self._result = ndimage.convolve(self._previous.result(),self._coeffs, mode=self._mode)
+
+'''
+ CorrelateCommand : Applies the supplied correlation filter
+'''
+class CorrelateCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._coeffs = jsonArgs.get('coeffs',[[0,0,1],[0,1,0],[1,0,0],[0,1,0],[0,0,1]])
+        self._mode = jsonArgs.get('mode','constant')
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Correlates the supplied filter with the data."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['coeffs','mode'] }
+        docs['args']['coeffs']['default'] = [[0,0,1],[0,1,0],[1,0,0],[0,1,0],[0,0,1]]
+        return docs
+
+    def execute(self):
+        self._result = ndimage.correlate(self._previous.result(),self._coeffs, mode=self._mode)
+
+
 
 '''
   MedianFilterCommand : applies a 2D square median filter using ndimage.median_filter()
@@ -623,6 +783,32 @@ class SumCommand(BasicCommand):
     def execute(self):
         self._result = extra_numpy.agnostic_sum(self._previous.result(), axis=self._axis, keepdims=True)
 
+'''
+ RollCommand : rolls the data along the requested axis using extra_numpy.agnostic_roll
+'''
+class RollCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._axis  = jsonArgs.get('axis',1)
+        self._shift = jsonArgs.get('window_length',1)
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Rolls the data along the specified axis, using numpy.roll()."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['axis','window_length'] }
+        docs['args']['axis']['default']=1
+        docs['args']['window_length']['default']=1
+        return docs
+
+    def isGPU(self):
+        return True
+
+    def execute(self):
+        self._result = extra_numpy.agnostic_roll(self._previous.result(), self._shift, axis=self._axis)
+
+
+
+
 
 '''
  RunningMeanCommand : A 1D running mean averaging filter, using the extra_numpy.running_mean()
@@ -694,43 +880,58 @@ class CurveFitCommand(BasicCommand):
 class ClipCommand(BasicCommand):
     def __init__(self,command, jsonArgs):
         super().__init__(command, jsonArgs)
+        self._dirname = jsonArgs.get('directory_out','NONE')
         self._xaxis = jsonArgs.get('xaxis',None)
         self._taxis = jsonArgs.get('taxis',None)            
         self._xmin = jsonArgs.get('xmin',numpy.min(self._xaxis))
         self._xmax = jsonArgs.get('xmax',numpy.max(self._xaxis))
         self._tmin = jsonArgs.get('tmin',numpy.min(self._taxis))
         self._tmax = jsonArgs.get('tmax',numpy.max(self._taxis))
+        self._double_ended = jsonArgs.get('double_ended',-1)
+        self._fname = str(int(self._tmin))
 
     def docs(self):
         docs={}
-        docs['one_liner']="Clip the data so that all subsequent operations operate on a small window."
-        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xaxis','taxis','xmin','xmax','tmin','tmax'] }
+        docs['one_liner']="Clip the data so that all subsequent operations operate on a small window. If directory_out is specified the new axes will also be created in the storage directory"
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['xaxis','taxis','xmin','xmax','tmin','tmax','directory_out','double_ended'] }
         return docs
 
     def isGPU(self):
         return True
 
+    def postcond(self):
+        return [directory_services.path_join(self._dirname,self._fname+'.npy')]
+
     def execute(self):
-        ixmin = 0
-        ixmax = 0
+        data = self._previous.result()
+        xend = int(self._xaxis.shape[0]*0.5)
+        print(xend)
+        if self._double_ended==0:
+            self._xaxis = self._xaxis[:xend]
+            data = data[:xend,:]
+        if self._double_ended==1:
+            self._xaxis = self._xaxis[xend:]
+            data = data[xend:,:]
+            self._xaxis = self._xaxis[::-1]
+            data = data[::-1,:]
+        ixmin = (numpy.abs(self._xaxis-self._xmin)).argmin()
+        ixmax = (numpy.abs(self._xaxis-self._xmax)).argmin()
         itmin = 0
         itmax = 0
-        for x in self._xaxis:
-            if self._xmin>x:
-                ixmin+=1
-            if self._xmax>x:
-                ixmax+=1
         if self._taxis is not None:
-            for t in self._taxis:
-                if self._tmin>t:
-                    itmin+=1
-                if self._tmax>t:
-                    itmax+=1
+            itmin = (numpy.abs(self._taxis-self._tmin)).argmin()
+            itmax = (numpy.abs(self._taxis-self._tmax)).argmin()
         else:
             itmin=0
-            itmax=self._previous.result().shape[1]
-        self._result = (self._previous.result())[ixmin:ixmax,itmin:itmax]
-
+            itmax=data.shape[1]
+        print('XMIN',self._xmin,self._xaxis[ixmin], ixmin)
+        print('XMAX',self._xmax,self._xaxis[ixmax], ixmax)
+        self._result = (data)[ixmin:ixmax,itmin:itmax]
+        dirname = self._dirname
+        if not dirname=='NONE':
+            io_helpers.numpy_out(dirname,self._fname,self._result)
+            io_helpers.numpy_out(dirname,'measured_depth',self._xaxis[ixmin:ixmax])
+            io_helpers.numpy_out(dirname,'time',self._taxis[itmin:itmax])
 '''
  DiffCommand : numerical differencing with a window_length offset
 '''
@@ -752,6 +953,42 @@ class DiffCommand(BasicCommand):
 
     def execute(self):
         self._result = extra_numpy.agnostic_diff(self._previous.result(),n=self._n, axis=self._axis)
+        # boundary condition
+        if self._axis==1:
+            self._result[:,-1]=self._result[:,-2]
+        if self._axis==0:
+            self._result[-1,:]=self._result[-2,:]
+
+'''
+ GradientCommand : numerical central differencing
+'''
+class GradientCommand(BasicCommand):
+    def __init__(self,command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._xaxis = jsonArgs['xaxis']
+        self._taxis = jsonArgs['taxis']
+        self._edge_order = jsonArgs.get('edge_order',1)
+        self._axis = jsonArgs.get('axis',None)
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Numercal gradient of the data via central differencing."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['edge_order','axis'] }
+        return docs
+
+    def isGPU(self):
+        return False
+
+    def execute(self):
+        if self._axis is not None:
+            if self._axis==1:
+                self._result = numpy.gradient(self._previous.result(),self._taxis,edge_order=self._edge_order, axis=1)
+            else:
+                self._result = numpy.gradient(self._previous.result(),self._xaxis,edge_order=self._edge_order, axis=0)
+        else:
+            self._result = numpy.gradient(self._previous.result(),[self._xaxis,self._taxis],edge_order=self._edge_order)
+
+                
 
 '''
  DownsampleCommand : index-based downsampling. Note there is no filtering, so for anti-alias
@@ -940,8 +1177,33 @@ class WienerCommand(BasicCommand):
 
 
     def execute(self):
+        print('wiener ['+str(self._xdir) + ', ' + str(self._tdir) + ']')
         self._result = signal.wiener(self._previous.result(), [self._xdir, self._tdir], self._noisePower)
-        
+
+
+'''
+ WriteNPYCommand : writes out the current datablock to NPY format
+'''
+class WriteNPYCommand(BasicCommand):
+    def __init__(self, command, jsonArgs):
+        super().__init__(command, jsonArgs)
+        self._dirname = jsonArgs.get('directory_out','NONE')
+        self._fname = jsonArgs.get('date_dir','NONE')
+
+    def docs(self):
+        docs={}
+        docs['one_liner']="Write the current state of the processed data to the npy format."
+        docs['args'] = { a_key: universal_arglist[a_key] for a_key in ['directory_out'] }
+        return docs
+
+    def postcond(self):
+        return [directory_services.path_join(self._dirname,self._fname+'.npy')]
+
+    def execute(self):
+        super().execute()
+        dirname = self._dirname
+        if not dirname=='NONE':
+            io_helpers.numpy_out(dirname,self._fname,self._previous.result())
 
 '''
  WriteWITSMLCommand : writes the results to the WITSML/FBE format.
@@ -959,7 +1221,7 @@ class WriteWITSMLCommand(BasicCommand):
         self._lowf = jsonArgs.get('low_freq',[0])
         self._highf = jsonArgs.get('high_freq',[1])
         self._labellist = jsonArgs.get('labels',[])
-        print(jsonArgs)
+        #print(jsonArgs)
 
     def postcond(self):
         return [directory_services.path_join(self._outdir,self._datedir+'.fbe')]
@@ -1216,6 +1478,7 @@ def KnownCommands(knownList):
     knownList['NONE']           = BasicCommand
     knownList['data']           = DataLoadCommand
     knownList['abs']            = AbsCommand
+    knownList['angle']          = AngleCommand
     knownList['add']            = AddCommand
     knownList['analytic_signal']= AnalyticSignalCommand
     knownList['argmax']         = ArgmaxCommand
@@ -1223,6 +1486,8 @@ def KnownCommands(knownList):
     knownList['butter']         = ButterCommand
     knownList['clip']           = ClipCommand
     knownList['conj']           = ConjCommand
+    knownList['convolve']       = ConvolveCommand
+    knownList['correlate']      = CorrelateCommand
     knownList['count_peaks']    = CountPeaksCommand
     knownList['destripe']       = DestripeCommand
     knownList['diff']           = DiffCommand
@@ -1233,6 +1498,7 @@ def KnownCommands(knownList):
     knownList['gather']         = GatherCommand
     knownList['gaussian']       = GaussianCommand
     knownList['geometric_mean'] = GeometricMeanCommand
+    knownList['gradient']       = GradientCommand
     knownList['harmonic_mean']  = HarmonicMeanCommand
     knownList['ifft']           = IFFTCommand
     knownList['keras']          = KerasCommand
@@ -1247,16 +1513,22 @@ def KnownCommands(knownList):
     knownList['peak_to_peak']   = Peak2PeakCommand
     knownList['rms_from_fft']   = RMSfromFFTCommand
     knownList['real']           = RealCommand
+    knownList['rescale']        = RescaleCommand
+    knownList['roll']           = RollCommand
     knownList['running_mean']   = RunningMeanCommand
+    knownList['sobel']          = SobelCommand
+    knownList['soft_threshold'] = SoftThresholdCommand
     knownList['sum']            = SumCommand
     knownList['skewness']       = SkewnessCommand
     knownList['sta_lta']        = StaLtaCommand
     knownList['std_dev']        = StdDevCommand
     knownList['to_gpu']         = ToGPUCommand
+    knownList['unwrap']         = UnwrapCommand
     knownList['up_wave']        = UpCommand
     knownList['velocity_map']   = VelocityMapCommand
     knownList['velocity_mask']  = VelocityMaskCommand
     knownList['wiener']         = WienerCommand
+    knownList['write_npy']      = WriteNPYCommand
     knownList['write_witsml']   = WriteWITSMLCommand
     return knownList    
 
