@@ -1,6 +1,7 @@
 from __future__ import division, absolute_import, print_function
 
 __all__ = ['witsmlfbe','io_helpers','h5_helpers',
-           'dirswap','sgy','directory_services']
+           'dirswap','sgy','directory_services',
+           'timeshift']
 
 
